#include "ppc.h"
#include "M33.h"

#include "framebuffer.h"

PPC::PPC(float hfov, int _w, int _h) {

	w = _w;
	h = _h;
	a = V3(1.0f, 0.0f, 0.0f);
	b = V3(0.0f, -1.0f, 0.0f);
	C = V3(0.0f, 0.0f, 0.0f);
	float hfovRad = hfov * 3.14159f / 180.0f;
	c = V3(-(float)w / 2.0f, (float)h / 2.0f, -(float)w/(2.0f*tanf(hfovRad/2.0f)));

}

int PPC::Project(V3 P, V3 &pP) {
	// 0 point behind head
	// 1 point in front of head

	M33 M;
	M.SetColumn(0, a);
	M.SetColumn(1, b);
	M.SetColumn(2, c);

	M33 Minv = M.Inverted();
	M33 tmp = M * Minv;
	V3 Q = Minv * (P - C);
	if (Q[2] <= 0.0f)
		return 0;

	pP = V3(Q[0] / Q[2], Q[1] / Q[2], 1.0f / Q[2]);
	return 1;

}



void PPC::Pan(float theta) {

	V3 aD = (b * -1.0f);
	a = a.RotateThisVectorAboutAxis(aD, theta);
	c = c.RotateThisVectorAboutAxis(aD, theta);

}

void PPC::Tilt(float theta) {

	V3 aD = a;
	b = b.RotateThisVectorAboutAxis(aD, theta);
	c = c.RotateThisVectorAboutAxis(aD, theta);

}

void PPC::Roll(float theta) {

	V3 aD = a^b;
	a = a.RotateThisVectorAboutAxis(aD, theta);
	b = b.RotateThisVectorAboutAxis(aD, theta);
	c = c.RotateThisVectorAboutAxis(aD, theta);

}

float PPC::GetF() {

	V3 vd = (a^b).UnitVector();
	return c*vd;

}

void PPC::SetPose(V3 newC, V3 LaP, V3 upgv) {

	V3 newa, newb, newc;

	V3 newvd = (LaP - newC).UnitVector();

	newa = (newvd ^ upgv).UnitVector();
	newb = (newvd ^ newa).UnitVector();
	float f = GetF();
	newc = newvd * f - newa * (float)(w / 2) - newb * (float)h / 2;
	
	a = newa;
	b = newb;
	c = newc;
	C = newC;

}

V3 PPC::GetVD() {

	return (a^b).UnitVector();

}

void PPC::SetInterpolated(PPC *ppc0, PPC *ppc1, int fi, int fN) {

	*this = *ppc0;

	float ff = (float)fi / (float)(fN - 1);
	V3 newC = ppc0->C + (ppc1->C - ppc0->C) * ff;
	V3 newvd = ppc0->GetVD() + (ppc1->GetVD() - ppc0->GetVD()) * ff;
	V3 LaP = newC + newvd * 100.0f;
	V3 upgv = (ppc0->b + (ppc1->b - ppc0->b) * ff)*-1.0f;

	SetPose(newC, LaP, upgv);

}

void PPC::Visualize(float visf, PPC *ppc, FrameBuffer *fb) {

	fb->Draw3DPoint(C, V3(0.0f, 0.0f, 0.0f), 7, ppc);

	float scf = visf / GetF();
	V3 visa = a * scf;
	V3 visb = b * scf;
	V3 visc = c * scf;

	fb->Draw3DPoint(C + visc, V3(1.0f, 0.0f, 0.0f), 5, ppc);
	fb->Draw3DPoint(C + visc + visa*(float)w, V3(0.0f, 1.0f, 0.0f), 5, ppc);
	fb->Draw3DPoint(C + visc + visb*(float)h, V3(0.0f, 0.0f, 1.0f), 5, ppc);

	fb->Draw3DSegment(C, V3(0.0f, 0.0f, 0.0f), C + visc, V3(0.0f, 0.0f, 0.0f), ppc);

	fb->Draw3DSegment(C + visc, V3(0.0f, 0.0f, 0.0f), C + visc + visa*(float)w, V3(0.0f, 0.0f, 0.0f), ppc);
	fb->Draw3DSegment(C + visc + visa*(float)w, V3(0.0f, 0.0f, 0.0f),
		C + visc + visa*(float)w + visb*(float)h, V3(0.0f, 0.0f, 0.0f),
		ppc);
	fb->Draw3DSegment(C + visc + visa*(float)w + visb*(float)h, V3(0.0f, 0.0f, 0.0f),
		C + visc + visb*(float)h, V3(0.0f, 0.0f, 0.0f),
		ppc);
	fb->Draw3DSegment(C + visc + visb*(float)h, V3(0.0f, 0.0f, 0.0f),
		C + visc, V3(0.0f, 0.0f, 0.0f),
		ppc);

}




